@@ -1,24 +1,26 @@
 '''
-block_cache.py - this file is part of S3QL (http://s3ql.googlecode.com)
+block_cache.py - this file is part of S3QL.
 
-Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>
 
-This program can be distributed under the terms of the GNU GPLv3.
+This work can be distributed under the terms of the GNU GPLv3.
 '''
 
-from .common import BUFSIZE
+from . import BUFSIZE
 from .database import NoSuchRowError
+from .backends.common import NoSuchObject
 from .multi_lock import MultiLock
 from .logging import logging # Ensure use of custom logger class
 from collections import OrderedDict
 from contextlib import contextmanager
 from llfuse import lock, lock_released
-from queue import Queue, Empty as QueueEmpty
+from queue import Queue, Empty as QueueEmpty, Full as QueueFull
 import os
 import hashlib
 import shutil
 import threading
 import time
+import sys
 
 # standard logger for this module
 log = logging.getLogger(__name__)
@@ -30,6 +32,14 @@ QuitSentinel = object()
 # be flushed
 FlushSentinel = object()
 
+class NoWorkerThreads(Exception):
+    '''
+    Raised when trying to enqueue an object, but there
+    are no active consumer threads.
+    '''
+
+    pass
+
 class Distributor(object):
     '''
     Distributes objects to consumers.
@@ -41,34 +51,56 @@ class Distributor(object):
         self.slot = None
         self.cv = threading.Condition()
 
-    def put(self, obj):
+        #: Number of threads waiting to consume an object
+        self.readers = 0
+
+    def put(self, obj, timeout=None):
         '''Offer *obj* for consumption
-        
-        The method blocks until another thread calls `get()` to consume
-        the object.
+
+        The method blocks until another thread calls `get()` to consume the
+        object.
+
+        Return `True` if the object was consumed, and `False` if *timeout* was
+        exceeded without any activity in the queue (this means an individual
+        invocation may wait for longer than *timeout* if objects from other
+        threads are being consumed).
         '''
 
         if obj is None:
             raise ValueError("Can't put None into Queue")
 
         with self.cv:
-            while self.slot is not None:
-                self.cv.wait()
+            # Wait until a thread is ready to read
+            while self.readers == 0 or self.slot is not None:
+                log.debug('waiting for reader..')
+                if not self.cv.wait(timeout):
+                    log.debug('timeout, returning')
+                    return False
+
+            log.debug('got reader, enqueueing %s', obj)
+            self.readers -= 1
+            assert self.slot is None
             self.slot = obj
-            self.cv.notify_all()
+            self.cv.notify_all() # notify readers
+
+        return True
 
     def get(self):
         '''Consume and return an object
-        
-        The method blocks until another thread offers an object
-        by calling the `put` method.
+
+        The method blocks until another thread offers an object by calling the
+        `put` method.
         '''
         with self.cv:
-            while not self.slot:
+            self.readers += 1
+            self.cv.notify_all()
+            while self.slot is None:
+                log.debug('waiting for writer..')
                 self.cv.wait()
             tmp = self.slot
             self.slot = None
             self.cv.notify_all()
+
         return tmp
 
 
@@ -97,21 +129,20 @@ class SimpleEvent(object):
         finally:
             self.__cond.release()
 
-    def wait(self):
+    def wait(self, timeout=None):
         self.__cond.acquire()
         try:
-            self.__cond.wait()
-            return
+            return self.__cond.wait(timeout)
         finally:
             self.__cond.release()
 
 
 class CacheEntry(object):
     """An element in the block cache
-    
+
     Attributes:
     -----------
-    
+
     :dirty:    entry has been changed since it was last uploaded.
     :size:     current file size
     :pos: current position in file
@@ -120,12 +151,12 @@ class CacheEntry(object):
     __slots__ = [ 'dirty', 'inode', 'blockno', 'last_access',
                   'size', 'pos', 'fh', 'removed' ]
 
-    def __init__(self, inode, blockno, filename):
+    def __init__(self, inode, blockno, filename, mode='w+b'):
         super().__init__()
         # Writing 100MB in 128k chunks takes 90ms unbuffered and
         # 116ms with 1 MB buffer. Reading time does not depend on
         # buffer size.
-        self.fh = open(filename, "w+b", 0)
+        self.fh = open(filename, mode, 0)
         self.dirty = False
         self.inode = inode
         self.blockno = blockno
@@ -180,7 +211,7 @@ class CacheDict(OrderedDict):
     An ordered dictionary designed to store CacheEntries.
 
     Attributes:
-    
+
     :max_size: maximum size to which cache can grow
     :max_entries: maximum number of entries in cache
     :size: current size of all entries together
@@ -194,7 +225,7 @@ class CacheDict(OrderedDict):
 
     def remove(self, key):
         '''Remove *key* from disk and cache, update size'''
-        
+
         el = self.pop(key)
         el.close()
         el.unlink()
@@ -206,16 +237,16 @@ class CacheDict(OrderedDict):
 
 class BlockCache(object):
     """Provides access to file blocks
-    
+
     This class manages access to file blocks. It takes care of creation,
     uploading, downloading and deduplication.
- 
+
     This class uses the llfuse global lock. Methods which release the lock have
     are marked as such in their docstring.
-    
+
     Attributes
     ----------
-    
+
     :path: where cached data is stored
     :cache: ordered dictionary of cache entries
     :mlock: MultiLock to synchronize access to objects and cache entries
@@ -243,12 +274,15 @@ class BlockCache(object):
         self.transfer_completed = SimpleEvent()
 
         # Will be initialized once threads are available
-        self.to_upload = None 
+        self.to_upload = None
         self.to_remove = None
-        
+
         if not os.path.exists(self.path):
             os.mkdir(self.path)
 
+        # Initialized fromt the outside to prevent cyclic dependency
+        self.fs = None
+
     def __len__(self):
         '''Get number of objects in cache'''
         return len(self.cache)
@@ -299,45 +333,60 @@ class BlockCache(object):
     def _unlock_entry(self, inode, blockno, release_global=False,
                       noerror=False):
         '''Release lock on cache entry'''
-        
+
         if release_global:
             with lock_released:
                 self.mlock.release((inode, blockno), noerror=noerror)
         else:
             self.mlock.release((inode, blockno), noerror=noerror)
-            
+
     def destroy(self):
         '''Clean up and stop worker threads
-        
+
         This method should be called without the global lock held.
         '''
 
-        log.debug('destroy(): clearing cache...')
-        with lock:
-            self.clear()
+        log.debug('Dropping cache...')
+        try:
+            with lock:
+                self.drop()
+        except NoWorkerThreads:
+            log.error('Unable to drop cache, no upload threads left alive')
 
-        for t in self.upload_threads:
-            self.to_upload.put(QuitSentinel)
+        # Signal termination to worker threads. If some of them
+        # terminated prematurely, continue gracefully.
+        log.debug('Signaling upload threads...')
+        try:
+            for t in self.upload_threads:
+                self._queue_upload(QuitSentinel)
+        except NoWorkerThreads:
+            pass
 
-        for t in self.removal_threads:
-            self.to_remove.put(QuitSentinel)
+        log.debug('Signaling removal threads...')
+        try:
+            for t in self.removal_threads:
+                self._queue_removal(QuitSentinel)
+        except NoWorkerThreads:
+            pass
 
-        log.debug('destroy(): waiting for upload threads...')
+        log.debug('waiting for upload threads...')
         for t in self.upload_threads:
             t.join()
-        
-        log.debug('destroy(): waiting for removal threads...')
+
+        log.debug('waiting for removal threads...')
         for t in self.removal_threads:
             t.join()
 
         assert len(self.in_transit) == 0
         try:
-            self.to_remove.get_nowait()
+            while self.to_remove.get_nowait() is QuitSentinel:
+                pass
         except QueueEmpty:
             pass
         else:
-            raise RuntimeError('Internal error: removal queue not empty.')
-        
+            log.error('Could not complete object removals, '
+                      'no removal threads left alive')
+
         self.to_upload = None
         self.to_remove = None
         self.upload_threads = None
@@ -347,7 +396,7 @@ class BlockCache(object):
 
         log.debug('cleanup done.')
 
-        
+
     def _upload_loop(self):
         '''Process upload queue'''
 
@@ -359,7 +408,7 @@ class BlockCache(object):
 
             self._do_upload(*tmp)
 
-            
+
     def _do_upload(self, el, obj_id):
         '''Upload object'''
 
@@ -380,8 +429,8 @@ class BlockCache(object):
                                                      % obj_id).get_obj_size()
                     time_ = time.time() - time_
                     rate = el.size / (1024 ** 2 * time_) if time_ != 0 else 0
-                    log.debug('_do_upload(%s): uploaded %d bytes in %.3f seconds, %.2f MiB/s',
-                              obj_id, el.size, time_, rate)
+                    log.debug('uploaded %d bytes in %.3f seconds, %.2f MiB/s',
+                              el.size, time_, rate)
                 else:
                     obj_size = backend.perform_write(do_write, 's3ql_data_%d'
                                                      % obj_id).get_obj_size()
@@ -390,6 +439,35 @@ class BlockCache(object):
                 self.db.execute('UPDATE objects SET size=? WHERE id=?', (obj_size, obj_id))
                 el.dirty = False
 
+        except Exception as exc:
+            log.debug('upload of %d failed: %s', obj_id, exc)
+            # At this point we have to remove references to this storage object
+            # from the objects and blocks table to prevent future cache elements
+            # to be de-duplicated against this (missing) one. However, this may
+            # already have happened during the attempted upload. The only way to
+            # avoid this problem is to insert the hash into the blocks table
+            # *after* successfull upload. But this would open a window without
+            # de-duplication just to handle the special case of an upload
+            # failing.
+            #
+            # On the other hand, we also want to prevent future deduplication
+            # against this block: otherwise the next attempt to upload the same
+            # cache element (by a different upload thread that has not
+            # encountered problems yet) is guaranteed to link against the
+            # non-existing block, and the data will be lost.
+            #
+            # Therefore, we just set the hash of the missing block to NULL,
+            # and rely on fsck to pick up the pieces. Note that we cannot
+            # delete the row from the blocks table, because the id will get
+            # assigned to a new block, so the inode_blocks entries will
+            # refer to incorrect data.
+            #
+
+            with lock:
+                self.db.execute('UPDATE blocks SET hash=NULL WHERE obj_id=?',
+                                (obj_id,))
+            raise
+
         finally:
             self.in_transit.remove(el)
             self._unlock_obj(obj_id)
@@ -399,40 +477,56 @@ class BlockCache(object):
 
     def wait(self):
         '''Wait until an object has been uploaded
-        
+
         If there are no objects in transit, return immediately. This method
         releases the global lock.
         '''
 
-        if not self.transfer_in_progress():
-            return
+        # Loop to avoid the race condition of a transfer terminating
+        # between the call to transfer_in_progress() and wait().
+        while True:
+            if not self.transfer_in_progress():
+                return
 
-        with lock_released:
-            self.transfer_completed.wait()
+            with lock_released:
+                if self.transfer_completed.wait(timeout=5):
+                    return
 
-    def upload(self, el):
-        '''Upload cache entry `el` asynchronously
-        
-        This method releases the global lock.
+    def upload_if_dirty(self, el):
+        '''Upload cache entry asynchronously
+
+        This method releases the global lock. Return True if the object
+        is actually scheduled for upload.
         '''
 
-        log.debug('upload(%s): start', el)
+        log.debug('started with %s', el)
+
+        if el in self.in_transit:
+            return True
+        elif not el.dirty:
+            return False
 
         # Calculate checksum
         with lock_released:
+            self._lock_entry(el.inode, el.blockno)
+            added_to_transit = False
             try:
-                self._lock_entry(el.inode, el.blockno)
-                assert el not in self.in_transit
+                if el is not self.cache.get((el.inode, el.blockno), None):
+                    log.debug('%s got removed while waiting for lock', el)
+                    self._unlock_entry(el.inode, el.blockno)
+                    return False
+                if el in self.in_transit:
+                    log.debug('%s already in transit', el)
+                    self._unlock_entry(el.inode, el.blockno)
+                    return True
                 if not el.dirty:
-                    log.debug('upload(%s): not dirty, returning', el)
+                    log.debug('no longer dirty, returning')
                     self._unlock_entry(el.inode, el.blockno)
-                    return
-                if (el.inode, el.blockno) not in self.cache:
-                    log.debug('%s removed while waiting for lock', el)
-                    self._unlock_entry(el.inode, el.blockno)
-                    return
-                
+                    return False
+
+                log.debug('uploading %s..', el)
                 self.in_transit.add(el)
+                added_to_transit = True
                 sha = hashlib.sha256()
                 el.seek(0)
                 while True:
@@ -442,10 +536,12 @@ class BlockCache(object):
                     sha.update(buf)
                 hash_ = sha.digest()
             except:
-                self.in_transit.discard(el)
+                if added_to_transit:
+                    self.in_transit.discard(el)
                 self._unlock_entry(el.inode, el.blockno)
                 raise
 
+        obj_lock_taken = False
         try:
             try:
                 old_block_id = self.db.get_val('SELECT block_id FROM inode_blocks '
@@ -460,26 +556,27 @@ class BlockCache(object):
             # No block with same hash
             except NoSuchRowError:
                 obj_id = self.db.rowid('INSERT INTO objects (refcount, size) VALUES(1, -1)')
-                log.debug('upload(%s): created new object %d', el, obj_id)
+                log.debug('created new object %d', obj_id)
                 block_id = self.db.rowid('INSERT INTO blocks (refcount, obj_id, hash, size) '
                                          'VALUES(?,?,?,?)', (1, obj_id, hash_, el.size))
-                log.debug('upload(%s): created new block %d', el, block_id)
-                log.debug('upload(%s): adding to upload queue', el)
+                log.debug('created new block %d', block_id)
+                log.debug('adding to upload queue')
 
                 # Note: we must finish all db transactions before adding to
                 # in_transit, otherwise commit() may return before all blocks
                 # are available in db.
                 self.db.execute('INSERT OR REPLACE INTO inode_blocks (block_id, inode, blockno) '
                                 'VALUES(?,?,?)', (block_id, el.inode, el.blockno))
-                
+
                 with lock_released:
                     self._lock_obj(obj_id)
-                    self.to_upload.put((el, obj_id))
+                    obj_lock_taken = True
+                    self._queue_upload((el, obj_id))
 
             # There is a block with the same hash
             else:
                 if old_block_id != block_id:
-                    log.debug('upload(%s): (re)linking to %d', el, block_id)
+                    log.debug('(re)linking to %d', block_id)
                     self.db.execute('UPDATE blocks SET refcount=refcount+1 WHERE id=?',
                                     (block_id,))
                     self.db.execute('INSERT OR REPLACE INTO inode_blocks (block_id, inode, blockno) '
@@ -488,26 +585,56 @@ class BlockCache(object):
                 el.dirty = False
                 self.in_transit.remove(el)
                 self._unlock_entry(el.inode, el.blockno, release_global=True)
-                
+
                 if old_block_id == block_id:
-                    log.debug('upload(%s): unchanged, block_id=%d', el, block_id)
-                    return
-                
+                    log.debug('unchanged, block_id=%d', block_id)
+                    return False
+
         except:
             self.in_transit.discard(el)
             with lock_released:
                 self._unlock_entry(el.inode, el.blockno, noerror=True)
-                self._unlock_obj(obj_id, noerror=True)
+                if obj_lock_taken:
+                    self._unlock_obj(obj_id)
             raise
 
-        # Check if we have to remove an old block
-        if not old_block_id:
-            log.debug('upload(%s): no old block, returning', el)
-            return
+        if old_block_id:
+            self._deref_block(old_block_id)
+        else:
+            log.debug('no old block')
 
-        self._deref_block(old_block_id)
+        return obj_lock_taken
+
+
+    def _queue_upload(self, obj):
+        '''Put *obj* into upload queue'''
+
+        while True:
+            if self.to_upload.put(obj, timeout=5):
+                return
+            for t in self.upload_threads:
+                if t.is_alive():
+                    break
+            else:
+                raise NoWorkerThreads('no upload threads')
+
+    def _queue_removal(self, obj):
+        '''Put *obj* into removal queue'''
+
+        while True:
+            try:
+                self.to_remove.put(obj, timeout=5)
+            except QueueFull:
+                pass
+            else:
+                return
+
+            for t in self.removal_threads:
+                if t.is_alive():
+                    break
+            else:
+                raise NoWorkerThreads('no removal threads')
 
-    
     def _deref_block(self, block_id):
         '''Decrease reference count for *block_id*
 
@@ -517,7 +644,7 @@ class BlockCache(object):
 
         This method releases the global lock.
         '''
-        
+
         refcount = self.db.get_val('SELECT refcount FROM blocks WHERE id=?', (block_id,))
         if refcount > 1:
             log.debug('decreased refcount for block: %d', block_id)
@@ -527,7 +654,8 @@ class BlockCache(object):
         log.debug('removing block %d', block_id)
         obj_id = self.db.get_val('SELECT obj_id FROM blocks WHERE id=?', (block_id,))
         self.db.execute('DELETE FROM blocks WHERE id=?', (block_id,))
-        refcount = self.db.get_val('SELECT refcount FROM objects WHERE id=?', (obj_id,))
+        (refcount, size) = self.db.get_row('SELECT refcount, size FROM objects WHERE id=?',
+                                           (obj_id,))
         if refcount > 1:
             log.debug('decreased refcount for obj: %d', obj_id)
             self.db.execute('UPDATE objects SET refcount=refcount-1 WHERE id=?',
@@ -541,11 +669,19 @@ class BlockCache(object):
         # transit itself. We can release it immediately after, because
         # the object is no longer in the database.
         log.debug('adding %d to removal queue', obj_id)
-        
+
         with lock_released:
             self._lock_obj(obj_id)
             self._unlock_obj(obj_id)
-            self.to_remove.put(obj_id)
+
+            if size == -1:
+                # size == -1 indicates that object has not yet been uploaded.
+                # However, since we just acquired a lock on the object, we know
+                # that the upload must have failed. Therefore, trying to remove
+                # this object would just give us another error.
+                return
+
+            self._queue_removal(obj_id)
 
 
     def transfer_in_progress(self):
@@ -568,10 +704,14 @@ class BlockCache(object):
             except QueueEmpty:
                 tmp = FlushSentinel
 
-            if (tmp is FlushSentinel or tmp is QuitSentinel) and ids:
+            if tmp in (FlushSentinel,QuitSentinel) and ids:
                 log.debug('removing: %s', ids)
-                with self.backend_pool() as backend:
-                    backend.delete_multi(['s3ql_data_%d' % i for i in ids])
+                try:
+                    with self.backend_pool() as backend:
+                        backend.delete_multi(['s3ql_data_%d' % i for i in ids])
+                except NoSuchObject:
+                    log.warning('Backend lost object s3ql_data_%d' % ids.pop(0))
+                    self.fs.failsafe = True
                 ids = []
             else:
                 ids.append(tmp)
@@ -579,22 +719,21 @@ class BlockCache(object):
             if tmp is QuitSentinel:
                 break
 
-
     @contextmanager
     def get(self, inode, blockno):
         """Get file handle for block `blockno` of `inode`
-        
+
         This method releases the global lock. The managed block, however,
         is executed with the global lock acquired and MUST NOT release
         it. This ensures that only one thread is accessing a given block
         at a time.
-        
+
         Note: if `get` and `remove` are called concurrently, then it is
         possible that a block that has been requested with `get` and
         passed to `remove` for deletion will not be deleted.
         """
 
-        #log.debug('get(inode=%d, block=%d): start', inode, blockno)
+        #log.debug('started with %d, %d', inode, blockno)
 
         if self.cache.is_full():
             self.expire()
@@ -612,15 +751,16 @@ class BlockCache(object):
                 self.cache.size += el.size - oldsize
         finally:
             self._unlock_entry(inode, blockno, release_global=True)
-            
-        #log.debug('get(inode=%d, block=%d): end', inode, blockno)
+
+        #log.debug('finished')
 
     def _get_entry(self, inode, blockno):
         '''Get cache entry for `blockno` of `inode`
 
         Assume that cache entry lock has been acquired.
         '''
-        
+
+        log.debug('started with %d, %d', inode, blockno)
         try:
             el = self.cache[(inode, blockno)]
 
@@ -633,21 +773,20 @@ class BlockCache(object):
 
             # No corresponding object
             except NoSuchRowError:
-                log.debug('_get_entry(inode=%d, block=%d): creating new block', inode, blockno)
+                log.debug('creating new block')
                 el = CacheEntry(inode, blockno, filename)
                 self.cache[(inode, blockno)] = el
                 return el
-            
+
             # Need to download corresponding object
             obj_id = self.db.get_val('SELECT obj_id FROM blocks WHERE id=?', (block_id,))
-            log.debug('_get_entry(inode=%d, block=%d): downloading object %d..',
-                      inode, blockno, obj_id)
-            el = CacheEntry(inode, blockno, filename)
-            try: 
+            log.debug('downloading object %d..', obj_id)
+            tmpfh = open(filename + '.tmp', 'wb')
+            try:
                 def do_read(fh):
-                    el.seek(0)
-                    el.truncate()
-                    shutil.copyfileobj(fh, el, BUFSIZE)
+                    tmpfh.seek(0)
+                    tmpfh.truncate()
+                    shutil.copyfileobj(fh, tmpfh, BUFSIZE)
 
                 with lock_released:
                     # Lock object. This ensures that we wait until the object
@@ -659,37 +798,42 @@ class BlockCache(object):
                     self._unlock_obj(obj_id)
                     with self.backend_pool() as backend:
                         backend.perform_read(do_read, 's3ql_data_%d' % obj_id)
+
+                tmpfh.flush()
+                os.fsync(tmpfh.fileno())
+                os.rename(tmpfh.name, filename)
             except:
-                el.unlink()
-                el.close()
+                os.unlink(tmpfh.name)
                 raise
-            
+            finally:
+                tmpfh.close()
+
+            el = CacheEntry(inode, blockno, filename, mode='r+b')
             self.cache[(inode, blockno)] = el
-            el.dirty = False # (writing will have set dirty flag)
             self.cache.size += el.size
 
         # In Cache
         else:
-            #log.debug('_get_entry(inode=%d, block=%d): in cache', inode, blockno)
+            #log.debug('in cache')
             self.cache.move_to_end((inode, blockno), last=True) # move to head
-            
+
         return el
 
     def expire(self):
         """Perform cache expiry
-        
+
         This method releases the global lock.
         """
 
         # Note that we have to make sure that the cache entry is written into
         # the database before we remove it from the cache!
 
-        log.debug('expire: start')
+        log.debug('started')
 
         while True:
             need_size = self.cache.size - self.cache.max_size
             need_entries = len(self.cache) - self.cache.max_entries
-            
+
             if need_size <= 0 and need_entries <= 0:
                 break
 
@@ -700,49 +844,46 @@ class BlockCache(object):
             for el in list(self.cache.values()):
                 if need_size <= 0 and need_entries <= 0:
                     break
-            
+
                 need_entries -= 1
                 need_size -= el.size
 
-                if el.dirty:
-                    if el not in self.in_transit:
-                        log.debug('expire: uploading %s..', el)
-                        self.upload(el) # Releases global lock
+                if self.upload_if_dirty(el): # Releases global lock
                     sth_in_transit = True
                     continue
 
-                log.debug('removing inode %d, block %d from cache', el.inode, el.blockno)
                 self._lock_entry(el.inode, el.blockno, release_global=True)
                 try:
                     # May have changed while we were waiting for lock
+                    if el is not self.cache.get((el.inode, el.blockno), None):
+                        log.debug('%s removed while waiting for lock', el)
+                        continue
                     if el.dirty:
                         log.debug('%s got dirty while waiting for lock', el)
                         continue
-                    if (el.inode, el.blockno) not in self.cache:
-                        log.debug('%s removed while waiting for lock', el)
-                        continue
+                    log.debug('removing %s from cache', el)
                     self.cache.remove((el.inode, el.blockno))
                 finally:
                     self._unlock_entry(el.inode, el.blockno, release_global=True)
 
             if sth_in_transit:
-                log.debug('expire: waiting for transfer threads..')
+                log.debug('waiting for transfer threads..')
                 self.wait() # Releases global lock
 
-        log.debug('expire: end')
+        log.debug('finished')
 
 
     def remove(self, inode, start_no, end_no=None):
         """Remove blocks for `inode`
-        
+
         If `end_no` is not specified, remove just the `start_no` block.
         Otherwise removes all blocks from `start_no` to, but not including,
-         `end_no`. 
-        
+         `end_no`.
+
         This method releases the global lock.
         """
 
-        log.debug('remove(inode=%d, start=%d, end=%s): start', inode, start_no, end_no)
+        log.debug('started with %d, %d, %s', inode, start_no, end_no)
 
         if end_no is None:
             end_no = start_no + 1
@@ -751,15 +892,14 @@ class BlockCache(object):
             self._lock_entry(inode, blockno, release_global=True)
             try:
                 if (inode, blockno) in self.cache:
-                    log.debug('remove(inode=%d, blockno=%d): removing from cache',
-                              inode, blockno)
+                    log.debug('removing from cache')
                     self.cache.remove((inode, blockno))
 
                 try:
                     block_id = self.db.get_val('SELECT block_id FROM inode_blocks '
                                                'WHERE inode=? AND blockno=?', (inode, blockno))
                 except NoSuchRowError:
-                    log.debug('remove(inode=%d, blockno=%d): block not in db', inode, blockno)
+                    log.debug('block not in db')
                     continue
 
                 # Detach inode from block
@@ -768,13 +908,13 @@ class BlockCache(object):
 
             finally:
                 self._unlock_entry(inode, blockno, release_global=True)
-                
+
             # Decrease block refcount
             self._deref_block(block_id)
 
-        log.debug('remove(inode=%d, start=%d, end=%s): end', inode, start_no, end_no)
+        log.debug('finished')
 
-    def flush(self, inode, blockno):
+    def flush_local(self, inode, blockno):
         """Flush buffers for given block"""
 
         try:
@@ -784,41 +924,104 @@ class BlockCache(object):
 
         el.flush()
 
-    def commit(self):
+    def start_flush(self):
         """Initiate upload of all dirty blocks
-        
+
         When the method returns, all blocks have been registered
-        in the database (but the actual uploads may still be 
+        in the database (but the actual uploads may still be
         in progress).
-        
+
         This method releases the global lock.
         """
 
-        # Need to make copy, since we aren't allowed to change dict while
-        # iterating through it. Look at the comments in CommitThread.run()
-        # (mount.py) for an estimate of the resulting performance hit.
+        # Need to make copy, since dict() may change while global lock is
+        # released. Look at the comments in CommitThread.run() (mount.py) for an
+        # estimate of the performance impact.
         for el in list(self.cache.values()):
-            if not el.dirty or el in self.in_transit:
-                continue
+            self.upload_if_dirty(el) # Releases global lock
+
+    def flush(self):
+        """Upload all dirty blocks
 
-            self.upload(el) # Releases global lock
-            
-    def clear(self):
-        """Clear cache
-        
         This method releases the global lock.
         """
 
-        log.debug('clear: start')
+        log.debug('started')
+
+        while True:
+            sth_in_transit = False
+
+            # Need to make copy, since dict() may change while global lock is
+            # released. Look at the comments in CommitThread.run() (mount.py)
+            # for an estimate of the performance impact.
+            for el in list(self.cache.values()):
+                if self.upload_if_dirty(el): # Releases global lock
+                    sth_in_transit = True
+
+            if not sth_in_transit:
+                break
+
+            log.debug('waiting for transfer threads..')
+            self.wait() # Releases global lock
+
+        log.debug('finished')
+
+    def drop(self):
+        """Drop cache
+
+        This method releases the global lock.
+        """
+
+        log.debug('started')
         bak = self.cache.max_entries
         self.cache.max_entries = 0
         self.expire() # Releases global lock
         self.cache.max_entries = bak
+        log.debug('finished')
 
-        log.debug('clear: end')
+    def get_usage(self):
+        '''Get cache usage information.
 
+        Return a tuple of
+
+        * cache entries
+        * cache size
+        * dirty cache entries
+        * dirty cache size
+        * pending removals
+
+        This method is O(n) in the number of cache entries.
+        '''
+
+        used = self.cache.size
+        dirty_size = 0
+        dirty_cnt = 0
+        for el in self.cache.values():
+            if el.dirty:
+                dirty_size += el.size
+                dirty_cnt += 1
+
+        if self.to_remove is None:
+            remove_cnt = 0
+        else:
+            # This is an estimate which may be negative
+            remove_cnt = max(0, self.to_remove.qsize())
+
+        return (len(self.cache), used, dirty_cnt, dirty_size, remove_cnt)
 
     def __del__(self):
-        if len(self.cache) > 0:
-            raise RuntimeError("BlockManager instance was destroyed without calling destroy()!")
+        # break reference loop
+        self.fs = None
 
+        if len(self.cache) == 0:
+            return
+
+        # Force execution of sys.excepthook (exceptions raised
+        # by __del__ are ignored)
+        try:
+            raise RuntimeError("BlockManager instance was destroyed without "
+                               "calling destroy()!")
+        except RuntimeError:
+            exc_info = sys.exc_info()
+
+        sys.excepthook(*exc_info)
