@@ -10,7 +10,8 @@ from .logging import logging, setup_logging, QuietError
 from . import fs, CURRENT_FS_REV
 from .backends.pool import BackendPool
 from .block_cache import BlockCache
-from .common import (get_seq_no, get_backend_factory, load_params, save_params)
+from .common import (get_seq_no, get_backend_factory, load_params, save_params,
+                     is_mounted)
 from .daemonize import daemonize
 from .database import Connection
 from .inode_cache import InodeCache
@@ -84,6 +85,12 @@ def main(args=None):
     if not os.path.exists(options.mountpoint):
         raise QuietError('Mountpoint does not exist.', exitcode=36)
 
+    # Check if fs is mounted on this computer
+    # This is not foolproof but should prevent common mistakes
+    if is_mounted(options.storage_url):
+        raise QuietError('File system already mounted elsewhere on this '
+                         'machine.', exitcode=40)
+
     if options.threads is None:
         options.threads = determine_threads(options)
 
@@ -740,11 +747,10 @@ class CommitThread(Thread):
                 # ... number=500)/500 * 1e3
                 # 1.456586996000624
                 for el in list(self.block_cache.cache.values()):
-                    if self.stop_event.is_set():
+                    if self.stop_event.is_set() or stamp - el.last_access < 10:
                         break
-                    if stamp - el.last_access < 10:
-                        break
-                    if self.block_cache.upload_if_dirty(el):
+                    if el.dirty and el not in self.block_cache.in_transit:
+                        self.block_cache.upload_if_dirty(el)
                         did_sth = True
 
             if not did_sth:
