@@ -1,10 +1,10 @@
 #!/usr/bin/env python3
 '''
-t4_adm.py - this file is part of S3QL (http://s3ql.googlecode.com)
+t4_adm.py - this file is part of S3QL.
 
-Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>
 
-This program can be distributed under the terms of the GNU GPLv3.
+This work can be distributed under the terms of the GNU GPLv3.
 '''
 
 if __name__ == '__main__':
@@ -13,20 +13,15 @@ if __name__ == '__main__':
     sys.exit(pytest.main([__file__] + sys.argv[1:]))
 
 from s3ql.backends import local
-from s3ql.backends.common import BetterBackend
+from s3ql.backends.comprenc import ComprencBackend
+from argparse import Namespace
 import shutil
-import sys
 import tempfile
 import unittest
 import subprocess
-import os.path
-
-if __name__ == '__main__':
-    mypath = sys.argv[0]
-else:
-    mypath = __file__
-BASEDIR = os.path.abspath(os.path.join(os.path.dirname(mypath), '..'))
+import pytest
 
+@pytest.mark.usefixtures('s3ql_cmd_argv', 'pass_reg_output')
 class AdmTests(unittest.TestCase):
 
     def setUp(self):
@@ -41,8 +36,8 @@ class AdmTests(unittest.TestCase):
         shutil.rmtree(self.backend_dir)
 
     def mkfs(self):
-        proc = subprocess.Popen([sys.executable, os.path.join(BASEDIR, 'bin', 'mkfs.s3ql'),
-                                 '-L', 'test fs', '--max-obj-size', '500', '--fatal-warnings',
+        proc = subprocess.Popen(self.s3ql_cmd_argv('mkfs.s3ql') +
+                                ['-L', 'test fs', '--max-obj-size', '500',
                                  '--authfile', '/dev/null', '--cachedir', self.cache_dir,
                                  '--quiet', self.storage_url ],
                                 stdin=subprocess.PIPE, universal_newlines=True)
@@ -52,15 +47,17 @@ class AdmTests(unittest.TestCase):
         proc.stdin.close()
 
         self.assertEqual(proc.wait(), 0)
+        self.reg_output(r'^WARNING: Maximum object sizes less than '
+                        '1 MiB will degrade performance\.$', count=1)
 
     def test_passphrase(self):
         self.mkfs()
 
         passphrase_new = 'sd982jhd'
 
-        proc = subprocess.Popen([sys.executable, os.path.join(BASEDIR, 'bin', 's3qladm'),
-                                 '--quiet', '--fatal-warnings', '--log', 'none', '--authfile',
-                                 '/dev/null', 'passphrase', self.storage_url ],
+        proc = subprocess.Popen(self.s3ql_cmd_argv('s3qladm') +
+                                [ '--quiet', '--log', 'none', '--authfile',
+                                  '/dev/null', 'passphrase', self.storage_url ],
                                 stdin=subprocess.PIPE, universal_newlines=True)
 
         print(self.passphrase, file=proc.stdin)
@@ -70,10 +67,11 @@ class AdmTests(unittest.TestCase):
 
         self.assertEqual(proc.wait(), 0)
 
-        plain_backend = local.Backend(self.storage_url, None, None)
-        backend = BetterBackend(passphrase_new.encode(), ('zlib', 6), plain_backend)
-        
-        backend.fetch('s3ql_passphrase') # will fail with wrong pw 
+        plain_backend = local.Backend(Namespace(
+            storage_url=self.storage_url))
+        backend = ComprencBackend(passphrase_new.encode(), ('zlib', 6), plain_backend)
+
+        backend.fetch('s3ql_passphrase') # will fail with wrong pw
 
 
     def test_authinfo(self):
@@ -90,9 +88,9 @@ class AdmTests(unittest.TestCase):
                   file=fh, sep='\n')
             fh.flush()
 
-            proc = subprocess.Popen([sys.executable, os.path.join(BASEDIR, 'bin', 'fsck.s3ql'),
-                                     '--quiet', '--fatal-warnings', '--authfile', fh.name,
-                                     '--cachedir', self.cache_dir, '--log', 'none', self.storage_url ],
+            proc = subprocess.Popen(self.s3ql_cmd_argv('fsck.s3ql') +
+                                    [ '--quiet', '--authfile', fh.name,
+                                      '--cachedir', self.cache_dir, '--log', 'none', self.storage_url ],
                                     stdin=subprocess.PIPE, universal_newlines=True)
 
             proc.stdin.close()
